/*
 *	---=== VERSION FOR 2.6.x KERNELS ===---
 */

/* SF64-PCR radio card (alias PCI RadioLINK) driver for Linux
 *
 * version 0.2.1, 14. 7. 2004 (modified to work with 2.6 kernel)
 *
 * (c) 2002 Martin "Marble" Beranek <marble@matfyz.cz>
 *
 ***************************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 * *************************************************************************
 *
 * Changelog:
 *
 * 0.2.1:
 * 	added PCR_MUTE to the pcr_cleanup_module() function;
 *
 * *************************************************************************
 *
 * some ideas from following sources
 *
 *  - SFR64-PCR FreeBSD driver by Vladimir Popov <jumbo@narod.ru>
 *  - SF-16FMR driver by David Sauer, davids@penguin.cz
 *  - original unusable driver :)
 *
 * TODO:
 *
 *  - to read 'tuned' info and real frequency
 *  - try to fix "self tuned radio" feature
 *
 * Notes:
 *
 *  - not tested on the other kernel ver. than 2.4.16
 *  - you have to enable 'Video for Linux'
 *
 */

#include <linux/module.h>	/* Modules 			*/
#include <linux/init.h>		/* Initdata			*/
#include <linux/ioport.h>	/* check_region, request_region	*/
#include <linux/delay.h>	/* udelay			*/
#include <asm/io.h>		/* outb, outb_p			*/
#include <asm/uaccess.h>	/* copy to/from user		*/
#include <linux/videodev.h>	/* kernel radio structs		*/
#include <linux/spinlock.h>
#include <linux/pci.h>


// here should be unique ID for this card...
#define VID_HARDWARE_SF64PCR	12345

#define SF64PCR_VENDOR		0x1319
#define SF64PCR_DEVICE		0x0801

struct pcr_device {
	long port;
	unsigned long freq;
	unsigned long lock, stereo, band, search;
        __u32 flags;
};

static int radio_nr = -1;
static spinlock_t lock;


/* referencnmi frekvence na karte (v kHz) */
#define SF64PCR_BASEFREQ	10700

/* from 1/16 kHz to the value for TEA5757 */
#define SF64PCR_ENCODE(x)	((((x)/16 + SF64PCR_BASEFREQ)*10 + 62)/125)

/* in 1/16 kHz */
#define SF64PCR_MINFREQ (80*16000)	/*  80MHz */
#define SF64PCR_MAXFREQ (110*16000)	/* 110MHz */

/****** TEA5757 SHIFT REGISTER ***********************************/

#define TEA5757_SEARCH_START		(1<<24) /* 0x1000000 */
#define TEA5757_SEARCH_END		(0<<24) /* 0x0000000 */

/* pcr_device->search */
#define TEA5757_SEARCH_UP		(1<<23) /* 0x0800000 */
#define TEA5757_SEARCH_DOWN		(0<<23) /* 0x0000000 */

/* pcr_device->stereo */
#define TEA5757_STEREO			(0<<22) /* 0x0000000 */
#define TEA5757_MONO			(1<<22) /* 0x0400000 */

/* pcr_device->band */
/* But no oscilator and other parts for AM receiving... :-( */
#define TEA5757_BAND_FM			(0<<20)
#define TEA5757_BAND_MW			(1<<20)
#define TEA5757_BAND_LW			(2<<20)
#define TEA5757_BAND_SW			(3<<20)

/* pcr_device->lock */
#define TEA5757_S005			(0<<16) /* 0x0000000 */ /* > 5 uV */
#define TEA5757_S010			(2<<16) /* 0x0010000 */ /* > 10 uV */
#define TEA5757_S030			(1<<16) /* 0x0020000 */ /* > 30 uV */
#define TEA5757_S150			(3<<16) /* 0x0030000 */ /* > 150 uV */

#define TEA5757_FREQ			0x0007FFF /* mask for freq. setting */

#define PCR_MUTE(dev) ((dev)->flags |= VIDEO_AUDIO_MUTE)
#define PCR_UNMUTE(dev) ((dev)->flags &= ~VIDEO_AUDIO_MUTE)
#define PCR_IS_MUTE(dev) ((dev)->flags & VIDEO_AUDIO_MUTE)



static void pcr_write(struct pcr_device* dev)
{
	int d, i;
	unsigned long freq = SF64PCR_ENCODE(dev->freq), reg = 0ul;


	freq &= TEA5757_FREQ;
	reg = dev->search | dev->band | dev->lock | dev->stereo | freq;

        spin_lock(&lock);
	outw(0xf800, dev->port);
	udelay(4);

	for(i=24; i>=0; i--) {
		if (reg & (1<<i))
			d = 0x4;
		else
			d = 0x0;
		outw(0xf800 + d, dev->port);
		udelay(4);
		outw(0xf801 + d, dev->port);
		udelay(4);
		outw(0xf800 + d, dev->port);
		udelay(4);
	}
        if PCR_IS_MUTE(dev)
		outw(0xf800, dev->port);
	else
		outw(0xf802, dev->port);
	udelay(4);
	spin_unlock(&lock);

	return;
}

static int pcr_getsigstr(struct pcr_device *dev)
{
	/* TODO */

	return 0xffff;
}


static int pcr_do_ioctl(struct inode *inode, struct file *file,
		unsigned int cmd, void *arg)
{
	struct video_device *dev = video_devdata(file);
	struct pcr_device *pcr = dev->priv;
	
	switch(cmd) {
	case VIDIOCGCAP:
	{
		struct video_capability *v = arg;

		memset(v,0,sizeof(*v));
		strcpy(v->name, "SF64-PCR radio");
		v->type=VID_TYPE_TUNER;
		v->channels=1;
		v->audios=0;	/* ??? I'm not sure, but hope that so. */
		return 0;
	}
	case VIDIOCGTUNER:
	{
		struct video_tuner *v = arg;

		if(v->tuner)	/* Only 1 tuner */
			return -EINVAL;
		strcpy(v->name, "FM");
		v->rangelow = SF64PCR_MINFREQ;
		v->rangehigh = SF64PCR_MAXFREQ;
		v->flags = pcr->flags &
			(VIDEO_TUNER_LOW | VIDEO_TUNER_STEREO_ON);
		v->mode = VIDEO_MODE_AUTO;
		v->signal = pcr_getsigstr(pcr);
		return 0;
	}
	case VIDIOCSTUNER:
	{
		struct video_tuner *v = arg;
		if(v->tuner!=0)
			return -EINVAL;
		return 0;
	}
	case VIDIOCGFREQ:
	{
		unsigned long *freq = arg;
		*freq = pcr->freq;
		return 0;
		return 0;
	}
	case VIDIOCSFREQ:
	{
		unsigned long *freq = arg;
		if (*freq<SF64PCR_MINFREQ || *freq>SF64PCR_MAXFREQ)
			return -EINVAL;
		pcr->freq = *freq;
		pcr_write(pcr);
		return 0;
	}
	case VIDIOCGAUDIO:
	{
		struct video_audio *v = arg;
		memset(v, 0, sizeof(*v));
		if (pcr->flags & VIDEO_AUDIO_MUTE)
			v->volume = 0;
		else
			v->volume = 0xFFFF;
		v->flags = (((pcr->flags & VIDEO_AUDIO_MUTE)?VIDEO_AUDIO_MUTE:0) |
				VIDEO_AUDIO_MUTABLE);
		strcpy(v->name, "Radio");
		v->mode = VIDEO_SOUND_STEREO;
		v->balance = 32768;
		v->step = 0xFFFF;
		return 0;			
	}
	case VIDIOCSAUDIO:
	{
		struct video_audio *v = arg;
		if(v->audio)
			return -EINVAL;
		if ((v->flags & VIDEO_AUDIO_MUTE) || (v->volume == 0))
			PCR_MUTE(pcr);
		else
			PCR_UNMUTE(pcr);
		pcr_write(pcr);
		return 0;
	}
        case VIDIOCGUNIT:
	{
		struct video_unit *v = arg;
		v->video=VIDEO_NO_UNIT;
		v->vbi=VIDEO_NO_UNIT;
		v->radio=dev->minor;
		v->audio=VIDEO_NO_UNIT;
		v->teletext=VIDEO_NO_UNIT;
		return 0;
	}
	default:
		return -ENOIOCTLCMD;
	}
}

static int pcr_ioctl(struct inode *inode, struct file *file,
		unsigned int cmd, unsigned long arg)
{
	return video_usercopy(inode, file, cmd, arg, pcr_do_ioctl);
}



static struct pcr_device pcr_unit;

static struct file_operations pcr_fops=
{
	.owner		= THIS_MODULE,
	.open           = video_exclusive_open,
	.release        = video_exclusive_release,
	.ioctl		= pcr_ioctl,
	.llseek		= no_llseek,
};


static struct video_device pcr_radio=
{
	.owner		= THIS_MODULE,
	.name		= "SF64PCR radio",
	.type		= VID_TYPE_TUNER,
	.hardware	= VID_HARDWARE_SF64PCR,
	.fops		= &pcr_fops,
};


static int __init pcr_init(void)
{
	struct pci_dev *dev = NULL;
	unsigned int port;
        
	dev=pci_find_device(SF64PCR_VENDOR, SF64PCR_DEVICE, dev);
	if (dev == NULL) {
		printk(KERN_ERR "sf64pcr: card not found.\n");
		return -1;
	}

	pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &port);
	port &= 0xFFFFFFF80;
	pcr_unit.port = port + 0x52;
	
	pcr_unit.freq = SF64PCR_MINFREQ;
	pcr_unit.stereo = TEA5757_STEREO;
	pcr_unit.lock = TEA5757_S030;
	pcr_unit.band = TEA5757_BAND_FM;
	pcr_unit.search = TEA5757_SEARCH_UP;
	pcr_unit.flags = VIDEO_TUNER_LOW | VIDEO_TUNER_STEREO_ON;
	PCR_MUTE(&pcr_unit);
        
	spin_lock_init(&lock);
	
	pcr_radio.priv = &pcr_unit;
	if (video_register_device(&pcr_radio, VFL_TYPE_RADIO, radio_nr) == -1)
		return -EINVAL;
        
        pcr_write(&pcr_unit);
	return 0;
}

void __exit pcr_cleanup_module(void)
{
	PCR_MUTE(&pcr_unit);
        video_unregister_device(&pcr_radio);
}


MODULE_AUTHOR("Martin \"Marble\" Beranek, marble@matfyz.cz");
MODULE_DESCRIPTION("Driver for the SF64-PCR (alias PCI RadioLINK) radio card.");
MODULE_LICENSE("GPL");

MODULE_PARM(radio_nr, "i");
MODULE_PARM_DESC(radio_nr, "Number of /dev/video[x] device.");

module_init(pcr_init)
module_exit(pcr_cleanup_module)

